import numpy as np
from ..genericas import print_verbose
from sympy import zeros, eye, simplify, sqrt


def permutacion_matriz(U, fila_i, idx_max, verbose=False, P=None, r=None):
    """Efectua una permutación por filas de una matriz

    Args:
        U (matriz): MAtriz a permutar
        fila_i (int): indice de fila origen
        idx_max (int): indice de fila a la que permutar
        verbose (bool, optional): verbose. Defaults to False.
        P (mat, optional): matriz de permutación. Defaults to None.
        r (mat, optional): rhs de ecuación. Defaults to None.

    Returns:
        (U, P, r): Matrices U, P y r cambiadas
    """
    print_verbose(
        [
            f"Permutamos fila {fila_i} con {idx_max}",
            f"U antes:\n {np.array(U)}",
            f"P antes:\n {np.array(P)}",
        ],
        verbose,
    )

    if fila_i != idx_max:
        fila_origen, fila_destino = U[fila_i, :].copy(), U[idx_max, :].copy()
        U[idx_max, :], U[fila_i, :] = fila_origen, fila_destino
        if P is not None:
            fila_origen, fila_destino = P[fila_i, :].copy(), P[idx_max, :].copy()
            P[idx_max, :], P[fila_i, :] = fila_origen, fila_destino
        if r is not None:
            fila_origen, fila_destino = r[fila_i, :].copy(), r[idx_max, :].copy()
            r[idx_max, :], r[fila_i, :] = fila_origen, fila_destino

    print_verbose(
        [f"U despues:\n {np.array(U)}", f"P despues:\n {np.array(P)}"], verbose
    )

    return U, P, r


def permutacion_L(L, perm, verbose=False):
    """
    Esta función la creo porque a la hora de hacer la permutación, hay que
    permutar los elementos de L pero no directamente. Solo hay que seleccionar
    los elementos de la diagonal inferior que correspondan con el menor número de
    columnas. Por ejemplo, si permutamos las filas 3 y 5, se tienen que mover solo
    los elementos de las 2 primeras columnas (2 -> 3) para luego continuar
    con las transformaciones del resto de columnas.

    Args:
        L (matriz): Matriz L de LDU
        perm (list): lista con indices inicial y final de permutación
        verbose (bool, optional): Verbose

    Returns:
        matriz: matriz L con la permutación aplicada correctamente a los elementos triangulares.
    """

    print_verbose(f"L antes:\n {np.array(L)}", verbose)

    fila_origen, fila_destino = (
        L[perm[0], : min(perm)].copy(),
        L[perm[1], : min(perm)].copy(),
    )
    L[perm[1], : min(perm)], L[perm[0], : min(perm)] = fila_origen, fila_destino

    print_verbose(f"L despues:\n {np.array(L)}", verbose)

    return L


def descomposicion_LU(m, rhs=None, verbose=True, permutar_max=False):
    """Esta función realiza el algoritmo de triangulación de Gauss.
        Para ello vamos a ir aplicando paso a paso el algoritmo tal
        cual se hace manualmente,
        y aplicamos los cambios de columnas necesarios si hay
        que aplicar permutaciones.
        Por defecto, si encontramos un 0 en la diagonal aplicamos
        permutar_max para esa
        fila, y devolvemos la matriz de permutaciones.

    Args:
        m (matriz): Matriz a descomponer
        rhs (matriz, optional): Lado derecho de la ecuación. Default es [1, 1, ..., 1]
        verbose (bool, optional): Aplicar verbose.
        permutar_max (bool, optional): Aplicar permutación por filas aun si no es necesario. Defaults to False.

    Returns:
        dict: {"P": matriz de permutaciones, "L": matriz L, U": matriz U, "r": matriz rhs adecuada a U}
    """

    if rhs is None:
        rhs = zeros(m.shape[0], 1)

    print_verbose(
        [
            "La matriz M|X es  (X = 0) si no se ha introducido",
            np.concatenate((np.array(m), np.array(rhs)), axis=1),
        ],
        verbose,
    )

    P, L, U, r = eye(m.shape[0]), eye(m.shape[0]), m.copy(), rhs.copy()
    fila_i, err, err_max = 0, 0, 3

    while fila_i < m.shape[0] and err < err_max:
        print_verbose(
            [
                f"\n=====================================\nFila {fila_i}",
                f"A {fila_i}, {err}",
            ],
            verbose,
        )

        if U[max(0, fila_i - 1), max(0, fila_i - 1)] == 0 or permutar_max:
            """
      Esta parte nos asegura que si un elemento diagonal es cero, permutamos
      la fila con su inmediata inferior y rehacemos los cálculos,
      y así tener una configuración viable.
      En cualquier permutación tenemos que cambiar la L acorde con el cambio.
      Al hacer la permutación, con los nuevos valores, repasamos la matriz para
      asegurarnos de que todos los puntos están cumplidos.

      Para la opción de permutar, buscamos un elemento de las filas no
      alteradas que sea el mayor. Si hay más de una fila, y una de
      ellas es la actual, no aplica la permutación.
      """
            sub_mat = U[fila_i:, fila_i:]
            max_el = np.max(np.array(sub_mat))
            idx_max = (
                np.min([i[0] for i in np.argwhere(np.array(sub_mat) == max_el)])
                + fila_i
            )  # Cogemos el primer elemento si hay varios

            if U[max(0, fila_i - 1), max(0, fila_i - 1)] == 0:
                U, P, r = permutacion_matriz(
                    U, max(0, fila_i - 1), idx_max, verbose, P, r
                )
                L = permutacion_L(L, [max(0, fila_i - 1), idx_max], verbose)
                fila_i = min(max(0, fila_i - 1), idx_max)

                err += 1
                print_verbose(f"Err {err}", verbose)
                continue

            elif permutar_max:
                if idx_max <= fila_i:  # No es necesario hacer la permutación
                    print_verbose(
                        "El índice de permutación es igual a la fila a permutar.",
                        verbose,
                    )
                else:
                    U, P, r = permutacion_matriz(U, fila_i, idx_max, verbose, P, r)
                    L = permutacion_L(L, [fila_i, idx_max], verbose)
                    continue

        # Ahora aplicamos el algoritmo de calculo de filas:
        for columna_j in range(fila_i):
            a_ij = U[fila_i, columna_j] / U[columna_j, columna_j]

            if a_ij != 0:
                L[fila_i, columna_j] = a_ij
                U[fila_i, :] = U[fila_i, :] - a_ij * U[columna_j, :]
                r[fila_i, :] = r[fila_i, :] - a_ij * r[columna_j, :]

            err = 0

            print_verbose(
                [
                    f"||||||||||||||||||||||||\n Columna {columna_j}",
                    f"a_{fila_i},{columna_j} = {a_ij}",
                    f"P = \n{np.array(P)}",
                    f"L = \n{np.array(L)}",
                    f"U = \n{np.array(U)}",
                    f"r = \n{np.array(r)}",
                ],
                verbose,
            )

        fila_i += 1

    if err == err_max:
        print("Algo ha ido mal... mira el log.")

    print_verbose(
        [
            "================= FORMA FINAL",
            f"a_{fila_i},{columna_j} = {a_ij}",
            f"P = \n{np.array(P)}\n",
            f"L = \n{np.array(L)}\n",
            f"U = \n{np.array(U)}\n",
            f"r = \n{np.array(r)}\n",
        ],
        verbose,
    )

    if simplify(L * U) != simplify(P * m):
        print("AVISO!!! LU != PA")

    return {"P": simplify(P), "L": simplify(L), "U": simplify(U), "r": simplify(r)}


def descomposicion_LDU(m, permutar_max=True, verbose=False):
    """Aplica la descomposición LDU = Pm a una matriz m.

    Args:
        m (matriz): matriz a permutar
        **args: args de descomposicion_LU

    Returns:
        dict: {"P": simplify(dict_LU["P"]),
        "L": simplify(L),
        "U": simplify(U),
        "D": simplify(D),}
    """
    dict_LU = descomposicion_LU(m, permutar_max=permutar_max, verbose=verbose)
    L, U = dict_LU["L"], dict_LU["U"]

    D = zeros(m.shape[0], m.shape[0])

    for i in range(U.shape[0]):
        D[i, i] = U[i, i]

    U = D.inv() * U

    print_verbose(
        [
            "================ RESULTADOS LDU*",
            f"L = \n{np.array(L)}\n",
            f"D = \n{np.array(D)}\n",
            f"U* = \n{np.array(U)}\n",
        ],
        verbose,
    )

    if simplify(L * D * U) != simplify(dict_LU["P"] * m):
        print("AVISO!!! LDU != PA")

    return {
        "P": simplify(dict_LU["P"]),
        "L": simplify(L),
        "U": simplify(U),
        "D": simplify(D),
    }


def cholesky(m, verbose=False):
    """
  Primero comprobamos que los menores sean positivos. Eso es equivalente que
  sus autovalores sean positivos. Por simplificar, si el determinante es negativo
  ya descartamos que sea factorizable, y saltamos el warning.
  """

    if m != m.T:
        print(
            "AVISO! La matriz no es simétrica, y por tanto no factorizable por Cholesky."
        )

    print_verbose(f"|M| es {m.det()}. Si es < 0, no es factorizable.", verbose)

    m_chol = zeros(m.shape[0], m.shape[1])

    for col_j in range(m.shape[0]):
        for row_i in range(col_j, m.shape[0]):
            if col_j == 0:
                if row_i == 0:
                    m_chol[row_i, col_j] = sqrt(m[row_i, col_j])
                else:
                    m_chol[row_i, col_j] = m[row_i, col_j] / m_chol[0, 0]

            else:
                if col_j == row_i:
                    m_chol[row_i, col_j] = sqrt(
                        m[col_j, col_j]
                        - sum([m_chol[col_j, k] ** 2 for k in range(col_j)])
                    )
                else:
                    m_chol[row_i, col_j] = (
                        m[row_i, col_j]
                        - sum(
                            [m_chol[col_j, k] * m_chol[row_i, k] for k in range(col_j)]
                        )
                    ) / (m_chol[col_j, col_j])

    if simplify(m_chol * m_chol.T) != m:
        print(
            f"AVISO!!! La matriz no es Cholesky-zable. \n L = \n {np.array(m_chol)}"
            "\n\n L*L.T = \n {np.array(m_chol * m_chol.T)}"
        )

    return simplify(m_chol)



import numpy as np
from sympy import zeros, eye, simplify, sqrt
from sympy import Matrix as mat

from ..genericas import print_verbose, norma
from .funcs_gram_schmidt import gram_schmidt


def householder(x, y, signo="+", normalizar=False):
    """Aplica la transformación de Householder. x e y pueden ser vectores fila o columna,
    en cuyo caso aplicamos la transformación de una forma u otra.

    Args:
        x (vector): vector origen
        y (vector): vector destino
        signo (str, optional): + o -, si - aplica la transformación inversa. Defaults to "+".
        normalizar (bool, optional): Normaliza los vectores para la transformación. Defaults to False.

    Returns:
        dict: {'H': la aplicación de transformación, 'e': el vector de eje de cambio.}
    """
    if signo == "+":
        mult = 1
    else:
        mult = -1

    if normalizar:
        x = simplify(x / norma(x))
        y = simplify(y / norma(y))

    if simplify(norma(x)) != simplify(norma(y)):
        print(
            f"AVISO!!! x tiene norma {norma(x)} e y tiene norma {norma(y)}. "
            "Para Householder las normas tienen que ser iguales."
        )

    if simplify(norma(x - y)) == 0:  # x==y
        e = zeros(x.shape[0], x.shape[1])
    else:
        e = mult * (x - y) / (norma(x - y))

    if (e.shape[0] == 1) and (e.shape[1] >= 1):  # El vector es fila:
        H = eye(x.shape[1]) - 2 * e.T * e
    else:
        H = eye(x.shape[0]) - 2 * e * e.T

    if (e.shape[0] == 1) and (e.shape[1] >= 1) and (simplify(x * H) != y):
        print(
            "AVISO FILAS!!! xH != y. Para Householder las normas tienen que ser iguales."
        )
    elif (e.shape[1] == 1) and (e.shape[0] >= 1) and (simplify(H * x) != y):
        print("AVISO!!! Hx != y. Para Householder las normas tienen que ser iguales.")

    return simplify(H), simplify(e)


def factorizacion_QR(m, verbose=True, metodo="gram_schmidt"):
    """Aplica el método QR para la factorización de la matriz m. m = QR, con Q ortogonal y R triangular superior.

    Args:
        m (matriz): matriz a factorizar.
        verbose (bool, optional): Imprime mensajes informativos. Defaults to True.
        metodo (str, optional): "gram_schmidt" o "householder".
        Householder funciona mejor para vectores proximos,
        donde gram-schmidt deja de ser estable. Defaults to "gram_schmidt".

    Returns:
        dict: {'Q', 'R' , 'D'} ('D' en gram-schmidt solo).
    """
    if metodo == "gram_schmidt":
        print_verbose("Aplicamos QR con Gram Schmidt", verbose)

        dict_gs = gram_schmidt(m)
        P, C = dict_gs["P"], dict_gs["c"]

        D = zeros(m.shape[0], m.shape[1])
        for col in range(m.shape[0]):
            D[col, col] = sqrt((P[:, col].T * P[:, col])[0])

        Q = simplify(P * (D ** (-1)))
        R = simplify(D * (eye(m.shape[0]) + C))

        print_verbose(
            f"Q es \n{np.array(Q)}\nR es \n{np.array(R)}\nD es \n{np.array(D)}", verbose
        )
    elif metodo == "householder":
        print_verbose("Aplicamos QR con Householder", verbose)
        lista_p = [eye(m.shape[0]) for _ in range(m.shape[0])]
        lista_e = []

        Q, R = eye(m.shape[0]), m.copy()
        for col in range(m.shape[1]):
            x = R[col:, col]
            y = mat([1] + [0] * (m.shape[0] - col - 1)) * norma(x)
            p_col = lista_p[col]
            H, e = householder(x, y)
            p_col[col:, col:] = H
            lista_e.append(e)

            R = simplify(p_col * R)
            Q = simplify(Q * p_col)

        print_verbose(f"Q es \n{np.array(Q)}\nR es \n{np.array(R)}", verbose)
    else:
        raise "El método no existe. Los disponibles son gram_schmidt y householder"

    if m != simplify(Q * R):
        print("AVISO!!! A != QR")

    if metodo == "gram_schmidt":
        return {"Q": Q, "R": R, "D": D}
    elif metodo == "householder":
        return {"Q": Q, "R": R}



from ..genericas import print_verbose
from sympy import zeros, simplify


def suma_columnas(lista):
    if len(lista) > 0:
        m = zeros(lista[0].shape[0], 1)
        for i in lista:
            m += i
        return m
    else:
        return 0


def gram_schmidt(m, verbose=False):
    """
  La ortogonalización produce una matriz ortogonalizada por columnas.
  p es la matriz ortogonal y p_norm es la ortonormal
  c es la matriz triangular tal que cij = aj·pi/||pi||^2 (los coeficientes de ortogonalización).
  Estos coeficientes se usan para la factorización QR.
  """

    p = zeros(m.shape[0], m.shape[1])

    p[:, 0] = m[:, 0]

    for col in range(1, m.shape[1]):
        p[:, col] = m[:, col] - suma_columnas(
            [
                (m[:, col].T * p[:, i])[0] / (p[:, i].T * p[:, i])[0] * p[:, i]
                for i in range(0, col)
            ]
        )

    print_verbose(f"La matriz ortogonal es \n {p}", verbose)

    p_norm = zeros(m.shape[0], m.shape[1])
    for col in range(p.shape[1]):
        p_norm[:, col] = p[:, col] / (p[:, col].T * p[:, col])[0]

    print_verbose(f"La matriz ortonormal es \n {p_norm}", verbose)

    c = zeros(m.shape[0], m.shape[1])
    for col in range(1, m.shape[1]):
        for row in range(0, col):
            c[row, col] = (m[:, col].T * p[:, row])[0] / (p[:, row].T * p[:, row])[0]

    return {"P": simplify(p), "Pn": simplify(p_norm), "c": simplify(c)}



from sympy import sqrt, simplify
import numpy as np


def norma(x):
    """Devuelve la norma de un vector <x,x>.

    Args:
        x (vector): Vector fila o columna

    Returns:
        float: Norma del vector
    """

    if x.shape[0] == 1:
        return sqrt((x * x.T)[0])
    else:
        return sqrt((x.T * x)[0])


def print_verbose(msg, verbose):
    """Printea el contenido de msg si verbose.

    Args:
        msg (str, list): mensaje a describir. Puede ser una list ade strings.
        verbose (bool): verbose
    """

    if type(msg) == str:
        msg = [msg]
    if verbose:
        for i in msg:
            print(i)


def norma_1(A, verbose=False):
    """Devuelve la norma 1, ||A||_1 de una matriz. (La suma de la columna con mayor suma).

    Args:
        A (matriz): Matriz
        verbose (bool, optional): Imprime cosas. Defaults to False.
    """
    norma_1 = max(np.sum(abs(A), axis=0))
    print_verbose(f"La norma 1 es: {norma_1}.", verbose)  # Mayor columna
    return norma_1


def norma_inf(A, verbose=False):
    """Devuelve la norma 1, ||A||_1 de una matriz. (La suma de la fila con mayor suma).

    Args:
        A (matriz): Matriz
        verbose (bool, optional): Imprime cosas. Defaults to False.
    """
    norma_inf = max(np.sum(abs(A), axis=1))
    print_verbose(f"La norma inf es: {norma_inf}.", verbose)  # Mayor fila
    return norma_inf


def norma_2(A, verbose=False):
    """Devuelve la norma 1, ||A||_1 de una matriz. Se define como la raíz del mayor de los autovalores de A*A.T.

    Args:
        A (matriz): Matriz
        verbose (bool, optional): Imprime cosas. Defaults to False.
    """
    eigs = sqrt([simplify(i) for i in list((A.T * A).eigenvals().keys())])
    norma_2 = np.max(eigs)
    print_verbose(f"Los eigs de A.T * A (sqrt) son: {eigs}", verbose)
    print_verbose(f"La norma 2 es: {norma_2}", verbose)
    return norma_2



from .funcs_LUD import descomposicion_LU, descomposicion_LDU, cholesky
from .funcs_gram_schmidt import gram_schmidt
from .funcs_householder_QR import householder, factorizacion_QR
from .funcs_met_iters import (
    metodo_iterativo,
    criterio_m_matriz,
    criterio_SOR,
    criterio_simetrica_definida_positiva,
    criterio_diagonal_dominante,
    criterio_radio_espectral,
)



from sympy import symbols, poly, zeros, simplify
from anmi.T2 import factorizacion_QR


def matriz_krylov(A, x, n_iters=None):
    """Genera una matriz de krylov dada una matriz A y un vector x. Cada columna de la matriz es la iteración i de A^i*x.

    Args:
        A (matriz): Matriz de aplicación
        x (vector): Vector base
        n_iters (int, optional): Número de iteraciones. Por defecto es el número de filas de A + 1 (garantiza que la matriz de krylov tiene una combinación lineal).

    Returns:
        m_krylov: Matriz con las aplicaciones de krylov por columna.
    """
    if n_iters is None:
        n_iters = A.shape[0] + 1

    m_krylov = zeros(A.shape[0], n_iters)
    m_krylov[:, 0] = x

    for i in range(1, n_iters):
        m_krylov[:, i] = A * m_krylov[:, i - 1]

    return simplify(m_krylov)


def sucesion_krylov(A, x):
    """
    La sucesión de krylov viene dada como la expresión (ejemplo para 3x3)
    p(l) = -l³ + a2 * l² + a1 * l + a0
    donde a = [a0, a1, a2] y es la solución del sistema de ecuaciones
    [x.T, Ax.T, Ax².T] * a.T = Ax³.T

    Args:
        A (matriz): Matriz para aplicar Krylov
        x (vector): Vector base

    Returns:
        poli_krylov (poly): Polinomio para la matriz de Krylov.
        a (vector): vector de coeficientes.
    """
    m_krylov = matriz_krylov(A, x)

    a = (m_krylov[:, : A.shape[0]] ** (-1)) * m_krylov[:, -1]
    a = simplify(a)

    lambda_ = symbols("lambda")
    poli_krylov = poly(-(lambda_ ** 3) + a[0])

    for i in range(1, A.shape[0]):
        poli_krylov += poly(a[i] * (lambda_ ** i))

    return poli_krylov, a


def potencia_iterada(A, x, n_iters, devolver_ultimo=True):
    """Aplica el método de la potencia iterada para calcular el mayor autovalor de la matriz, usando el método de Krylov.

    Args:
        A (matriz): Matriz aplicación
        x (vector): Vector base para el método. Si el vector es autovector dará fallo.
        n_iters (int): Número de iteraciones
        devolver_ultimo (bool, optional): Si True, devuelve el vector de la última iteración. Si False, devuelve todas las iteraciones.

    Returns:
        m_cocientes (matriz, vector): matriz/vector con el número de filas igual al de A, con los cocientes. Los números deberían tender al mayor autovalor de A.
    """

    m_krylov = matriz_krylov(A, x, n_iters=n_iters)

    if devolver_ultimo:
        m_cocientes = zeros(m_krylov.shape[0], 1)
        for row in range(m_cocientes.shape[0]):
            m_cocientes[row, 0] = (
                m_krylov[row, m_krylov.shape[1] - 1]
                / m_krylov[row, m_krylov.shape[1] - 2]
            )

    else:
        m_cocientes = zeros(m_krylov.shape[0], m_krylov.shape[1] - 1)
        for col in range(m_cocientes.shape[1]):
            for row in range(
                m_cocientes.shape[0]
            ):  # No hay element-wise division en sympy
                m_cocientes[row, col] = m_krylov[row, col + 1] / m_krylov[row, col]

    m_cocientes = simplify(m_cocientes)
    return m_cocientes


def metodo_autovals_QR(A, n_iters=3, verbose=True):
    """Aplica el método QR para el cálculo de autovalores de una matriz.

    Args:
        A (matriz): Matriz para el metodo
        n_iters (int, optional): Número de iteraciones. Defaults to 3.
        verbose (bool, optional): Imprime información intermedia.

    Returns:
        dict: "A": lista de valores de A = R*Q en cada iteración, "R" Y "Q": matrices Q y R del método.
    """
    list_A, list_Q, list_R = [], [], []

    # La lista de A tiene los A^{(2)} hasta A^{(k+1)}, porque A^{(1)} = A, y ya lo conocemos
    for i in range(n_iters):
        dict_QR = factorizacion_QR(A, verbose=verbose)
        Q, R = simplify(dict_QR["Q"]), simplify(dict_QR["R"])
        list_Q.append(Q)
        list_R.append(R)

        A = simplify(R * Q).copy()
        list_A.append(A)

    return {"A": list_A, "Q": list_Q, "R": list_R}



import setuptools


setuptools.setup(
    name="anmi",  # Replace with your own username
    version="0.0.2",
    author="alexmascension",
    author_email="alexmascension@gmail.com",
    description="Archivos para correr ANMI",
    long_description="",
    long_description_content_type="text/markdown",
    url="https://github.com/alexmascension/ANMI",
    project_urls={"Bug Tracker": "https://github.com/alexmascension/ANMI/issues",},
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: BSD License",
        "Operating System :: OS Independent",
    ],
    package_dir={"": "src"},
    packages=setuptools.find_packages(where="src"),
    python_requires=">=3.6",
)



from sympy import simplify, zeros
from sympy import Matrix as mat
import numpy as np

from ..genericas import print_verbose


def criterio_radio_espectral(H, verbose=True):
    eigs = [simplify(i) for i in list(H.eigenvals().keys())]

    print_verbose("||Criterio de radio espectral||", verbose)
    try:
        print_verbose(
            f"El mayor autovalor es {np.max(np.array(eigs, dtype=float))}. Si ese valor es < 1 entonces los métodos iterativos convergen.",
            verbose,
        )
    except:
        print_verbose(
            f"Los autovalores son {eigs}. Si el mayor autovalor es < 1, entonces el método converge.",
            verbose,
        )


def criterio_diagonal_dominante(A, verbose=True):
    print_verbose(
        "||Criterio de Diagonal Dominante||\n Si la matriz es dominante por filas, los métodos de Jacobi y Gauss-Seidel convergen.",
        verbose,
    )
    A_abs = abs(A)
    try:
        np.array(A_abs, dtype=float)
        for r in range(A.shape[0]):
            diff = 2 * A_abs[r, r] - sum(A_abs[r, :])
            if diff <= 0:
                print_verbose(
                    f"La fila {r} NO es dominante por filas: diff = {diff}.", verbose
                )
                return
        print_verbose("La matriz CUMPLE EL CRITERIO DIAGONAL DOMINANTE", verbose)
    except:
        print_verbose(
            "La matriz tiene complejos o simbolos. Hay que verificar el criterio a mano.",
            verbose,
        )


def criterio_simetrica_definida_positiva(A, verbose=True):
    print_verbose(
        "||Criterio de Sim Def Pos||\n Si la matriz es simétrica y definida positiva, el método de Gauss-Seidel es convergente.",
        verbose,
    )
    if A != A.T:
        print_verbose("La matriz NO es simétrica.", verbose)
        return

    det_A = A.det()
    print_verbose(f"El determinante de A es {det_A}.", verbose)
    try:
        if float(det_A) > 0:
            print_verbose(
                "La matriz es DEFINIDA POSITIVA (el determinante es positivo).",
                verbose,
            )
            print_verbose("La matriz CUMPLE EL CRITERIO SIM DEF POS", verbose)
        else:
            print_verbose(
                "La matriz NO es DEFINIDA POSITIVA (el determinante no es positivo).",
                verbose,
            )
    except:
        print_verbose(
            "No podemos determinar la positividad porque hay símbolos o complejos.",
            verbose,
        )


def criterio_SOR(verbose):
    print_verbose(
        "||Criterio SOR||\n Si la matriz es simétrica y definida positiva y w in (0, 2) el método SOR es convergente.\nSi w no (0, 2) el método SOR no converge.",
        verbose,
    )


def criterio_m_matriz(A, verbose):
    print_verbose(
        "||Criterio M matriz||\n Si la A es M-matriz entonces las descomposiciones de Jacobi y Gauss-Seidel son convergentes.\nA^-1 >= 0\naij < 0 para todo i =/= j",
        verbose,
    )
    A_inv = A ** -1

    try:
        np.array(A, dtype=float)
        if np.min(A_inv) >= 0:
            print_verbose("A^-1 >= 0", verbose)
        else:
            print_verbose("A^-1 < 0. La matriz NO CUMPLE el criterio", verbose)

        A_null_diag = A.copy()
        for i in range(A.shape[0]):
            A_null_diag[i, i] = 0

        if np.max(A_null_diag) > 0:
            print_verbose(
                "La matriz tiene elementos no diagonales positivos. NO CUMPLE el criterio.",
                verbose,
            )
        else:
            print_verbose("Los elementos no diagonales son negativos.", verbose)
    except:
        print_verbose(
            "La matriz tiene complejos o símbolos, no podemos verificar le criterio.",
            verbose,
        )


def metodo_iterativo(
    A, b=None, x0=None, metodo="jacobi", w=1.5, n_iter=10, verbose=True,
):
    """Aplica el método iterativo designado

    Args:
        A (matriz): Matriz de valores
        b (vector, optional): Vector de rhs. Por defecto es 1, 1, ..., 1.
        x0 (vector, optional): Vector con elementos de la primera iteración. Por defecto es 1, 1, ..., 1.
        metodo (str, optional): método de resolución, puede ser "jacobi", "gs" o "sor".
        w (float, optional): Peso para método sor. Defaults to 1.5.
        n_iter (int, optional): Número de iteraciones del método. Defaults to 10.
        verbose (bool, optional): Imprime resultados intermedios. Defaults to True.

    Returns:
        dict: 'x': vector de resultados para Ax=b, 'diff': diferencia entre Ax y b para cada iteración.
    """
    if b is None:
        b = mat([[1] * A.shape[0]]).T
    if x0 is None:
        x0 = mat([[1] * A.shape[1]]).T

    D, L, U = (
        zeros(A.shape[0], A.shape[1]),
        zeros(A.shape[0], A.shape[1]),
        zeros(A.shape[0], A.shape[1]),
    )
    for r in range(A.shape[0]):
        for c in range(A.shape[1]):
            if r == c:
                D[r, c] = A[r, c]
            elif r < c:
                U[r, c] = -A[r, c]
            else:
                L[r, c] = -A[r, c]

    if metodo == "jacobi":
        M = D
    elif metodo == "gs":
        M = D - L
    elif metodo == "sor":
        M = D / w - L

    N = simplify(M - A)

    # Aplicamos criterios!
    criterio_radio_espectral(M ** (-1) * N, verbose)
    criterio_diagonal_dominante(A, verbose)
    criterio_simetrica_definida_positiva(A, verbose)
    criterio_SOR(verbose)
    criterio_m_matriz(A, verbose)

    diff = []
    for iter in range(n_iter):  # Aplica el método
        x0 = (M ** (-1)) * (N * x0 + b)
        diff.append(np.sum(np.abs(A * x0 - b)))

    return {"x": x0, "diff": diff}
